 /*
  * Copyright (C), 2018-2021
  * Project: starter
  * FileName: LoginFilter
  * Author:   Louis
  * Date:     2021/3/30 22:09
  */
 package com.fr.plugin.xxxx.saml.request;

 import com.fanruan.api.decision.login.LoginKit;
 import com.fanruan.api.decision.user.UserKit;
 import com.fanruan.api.i18n.I18nKit;
 import com.fanruan.api.log.LogKit;
 import com.fanruan.api.net.NetworkKit;
 import com.fanruan.api.util.RenderKit;
 import com.fanruan.api.util.StringKit;
 import com.fr.decision.fun.impl.AbstractGlobalRequestFilterProvider;
 import com.fr.decision.webservice.utils.DecisionServiceConstants;
 import com.fr.decision.webservice.v10.login.LoginService;
 import com.fr.general.ComparatorUtils;
 import com.fr.plugin.context.PluginContexts;
 import com.fr.plugin.xxxx.saml.config.SamlConfig;
 import com.fr.plugin.xxxx.saml.util.RSA;
 import com.fr.stable.CodeUtils;
 import com.fr.stable.fun.Authorize;
 import com.fr.third.org.apache.http.NameValuePair;
 import com.fr.third.org.apache.http.client.utils.URIBuilder;
 import com.fr.web.utils.WebUtils;

 import javax.servlet.FilterChain;
 import javax.servlet.FilterConfig;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
 import java.net.URISyntaxException;
 import java.security.NoSuchAlgorithmException;
 import java.security.spec.InvalidKeySpecException;
 import java.util.HashMap;
 import java.util.List;
 import java.util.Map;

 import static com.fr.plugin.xxxx.saml.config.SamlConfig.*;

 /**
  * <Function Description><br>
  * <LoginFilter>
  *
  * @author fr.open
  * @since 1.0.0
  */
 @Authorize(callSignKey = PLUGIN_ID)
 public class LoginFilter extends AbstractGlobalRequestFilterProvider {
     public static final String STATE = "state";
     public static final String SAML_STATE = "saml";
     public static final int EXPIRE_TIME = 3000;

     private SamlConfig config;


     /**
      * 过滤器名称
      *
      * @return
      */
     @Override
     public String filterName() {
         return "SamlFilter";
     }

     /**
      * 过滤规则
      *
      * @return
      */
     @Override
     public String[] urlPatterns() {
         return new String[]{"/decision", "/decision/view/form", "/decision/view/report", "/decision/v10/entry/access/*"};
     }

     /**
      * 过滤器初始化
      *
      * @param filterConfig
      */
     @Override
     public void init(FilterConfig filterConfig) {
         this.config = SamlConfig.getInstance();
         super.init(filterConfig);
     }

     /**
      * 过滤器处理
      *
      * @param request
      * @param response
      * @param filterChain
      */
     @Override
     public void doFilter(HttpServletRequest request, HttpServletResponse response, FilterChain filterChain) {
         try {
             if (operation(request, response)) {
                 filterChain.doFilter(request, response);
             }
         } catch (Exception e) {
             LogKit.error(e.getMessage(), e);
         }
     }

     /**
      * 用户验证登陆操作
      *
      * @param req
      * @param res
      * @throws Exception
      */
     private boolean operation(HttpServletRequest req, HttpServletResponse res) throws Exception {
         String state = NetworkKit.getHTTPRequestParameter(req, STATE);
         String token = NetworkKit.getHTTPRequestParameter(req, SP_TOKEN);
         // 已登录
         if (LoginService.getInstance().isLogged(req)) {
             return true;
         }
         if (StringKit.isEmpty(token)) {
             redirectToLoginPage(req, res);
             return false;
         }
         if (StringKit.isBlank(state) || !StringKit.equals(state, SAML_STATE)) {
             return true;
         }
         String username = getUsername(token);
         if (StringKit.isEmpty(username) || !UserKit.existUsername(username)) {
             return true;
         }
         if (!PluginContexts.currentContext().isAvailable()) {
             LogKit.error(I18nKit.getLocText("Plugin-Saml_Licence_Expired"));
             return true;
         }
         String tokenFR = LoginKit.login(req, res, username);
         req.setAttribute(DecisionServiceConstants.FINE_AUTH_TOKEN_NAME, tokenFR);
         res.sendRedirect(removeToken(req));
         return false;
     }

     /**
      * 移除SAMl的token参数
      *
      * @param request
      * @return
      * @throws URISyntaxException
      */
     private String removeToken(HttpServletRequest request) throws URISyntaxException {
         URIBuilder uriBuilder = new URIBuilder(WebUtils.getOriginalURL(request).replaceAll(" ", "%2B"));
         List<NameValuePair> params = uriBuilder.getQueryParams();
         params.removeIf(pair -> ComparatorUtils.equals(pair.getName(), SP_TOKEN));
         params.removeIf(pair -> ComparatorUtils.equals(pair.getName(), STATE));
         uriBuilder.clearParameters();
         if (!params.isEmpty()) {
             uriBuilder.setParameters(params);
         }
         return uriBuilder.build().toString();
     }

     /**
      * 跳转登陆地址
      *
      * @param res
      * @return
      * @throws Exception
      */
     private void redirectToLoginPage(HttpServletRequest req, HttpServletResponse res) throws Exception {
         Map<String, Object> params = new HashMap<>();
         params.put("spUrl", this.config.getSpUrl());
         params.put("idpEntityId", CodeUtils.encodeURIComponent(this.config.getIdpEntityId()));
         params.put("redirectUri", CodeUtils.encodeURIComponent(WebUtils.getOriginalURL(req)));
         String loginUrl = RenderKit.renderParameter4Tpl(URL_SP, params);
         LogKit.info("saml-LoginFilter-redirectToLoginPage-loginUrl:{}", loginUrl);
         res.sendRedirect(loginUrl);
     }

     /**
      * 通过code获得用户名
      * 返回内容格式：时间戳#用户主体名称
      *
      * @param token
      * @return
      * @throws Exception
      */
     private String getUsername(String token) throws Exception {
         LogKit.info("saml-LoginFilter-operation-token:{}", token);
         String decryptData = RSA.privateDecrypt(token, RSA.getPrivateKey(PRIVATE_KEY));
         LogKit.info("saml-LoginFilter-operation-decryptData:{}", decryptData);
         long timeStamp = Long.parseLong(decryptData.substring(0, decryptData.indexOf("#")));
         String username = decryptData.substring(decryptData.indexOf("#") + 1);
         if (System.currentTimeMillis() - timeStamp > EXPIRE_TIME) {
             LogKit.error("saml-LoginFilter-getUsername-TimeStamp validation failed.");
             return StringKit.EMPTY;
         }
         LogKit.info("saml-LoginFilter-getUsername-username:{}", username);
         return username;
     }
 }